# Generated by Django 4.0.10 on 2023-10-14 13:07

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("memberaudit", "0013_add_corporation_titles"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="mailentity",
            options={"default_permissions": ()},
        ),
        migrations.AlterModelOptions(
            name="skillset",
            options={"verbose_name": "skill set", "verbose_name_plural": "skill sets"},
        ),
        migrations.AlterModelOptions(
            name="skillsetgroup",
            options={
                "verbose_name": "skill set group",
                "verbose_name_plural": "skill set groups",
            },
        ),
        migrations.AlterModelOptions(
            name="skillsetskill",
            options={"verbose_name": "skill", "verbose_name_plural": "skills"},
        ),
    ]
