{% load i18n %}
{% load evelinks %}

<img height="20" width="20" src="{{ character.character_id|character_portrait_url:32 }}">
<span class="sidebar-character-list-item-name" {% if not character.memberaudit_character_pk %}class="text-muted"{% endif%}>
    &nbsp;{{ character.character_name }}
</span>

{% if character.character_id == main_character_id %}
    &nbsp;<i class="sidebar-character-list-item-icon fas fa-crown" title="{% translate 'Main character' %}"></i>
{% endif %}

{% if character.is_shared %}
    &nbsp;<i class="sidebar-character-list-item-icon far fa-eye" title="{% translate 'Currently shared with recruiters' %}"></i>
{% endif %}
